//#define DEBUG_ASCII
//#define DEBUG_FIRMATA

// ASCII CODES
#define ASCII_SEPARATOR 32

// FIRMATA CODES
#define FIRMATA_MAJOR_VERSION   2 // for non-compatible changes
#define FIRMATA_MINOR_VERSION   1 // for backwards compatible changes
// message command bytes (128-255/0x80-0xFF)
#define DIGITAL_MESSAGE         0x90 // send data for a digital pin
#define ANALOG_MESSAGE          0xE0 // send data for an analog pin (or PWM)
#define REPORT_ANALOG           0xC0 // enable analog input by pin #
#define REPORT_DIGITAL          0xD0 // enable digital input by port pair
//
#define SET_PIN_MODE            0xF4 // set a pin to INPUT/OUTPUT
//
#define REPORT_VERSION          0xF9 // report protocol version
#define SYSTEM_RESET            0xFF // reset from MIDI
//
#define START_SYSEX             0xF0 // start a MIDI Sysex message
#define END_SYSEX               0xF7 // end a MIDI Sysex message



#include "WProgram.h"
#include "HardwareSerial.h"
#include "Tata.h"

// ADDED FOR COMPATIBILITY WITH WIRING
extern "C" {
#include <string.h>
#include <stdlib.h>
}


Tata_::Tata_() 
{
	callback = NULL;
	buffer[TATA_BUFFER_SAFE_NULL]=0;
	
	
	// Reporting
	millisToWait=millis();
    
	
    clear();
    
}

// Tata should usually be started at 57600
void Tata_::begin(long baud) 
{
	begin(baud,NULL,"basic");
	
}

// Tata should usually be started at 57600
void Tata_::begin(long baud,TataCallbackFunction newFunction,char* string) 
{
	#if defined(__AVR_ATmega128__)  // Wiring
		Serial.begin((uint32_t)baud);
	#else
		Serial.begin(baud);
	#endif
	callback = newFunction;
	description = string;
	
	// Attempt to force dump any leftover messages from the client software
	Serial.println();
	Serial.println();
	Serial.println();
	
	// Anounce that Tata is ready
	printVersion();
}


void Tata_::printVersion() {
	
	Serial.print("tata arduino ");
	Serial.println(description);
	
}

void Tata_::printVersionFirmata(void)
{
	
  Serial.print(REPORT_VERSION, BYTE);
  Serial.print(FIRMATA_MAJOR_VERSION, BYTE);
  Serial.print(FIRMATA_MINOR_VERSION, BYTE);

}


void Tata_::clear() {
	
	bufferIndex = 0;
     current = 0;
     message_ready = 0;
     valid_char = 0;
     word_count = 0;
     message_type = ASCII;
    
}

/*
uint8_t Tata_::type() {
  return message_type;

}
*/
uint8_t Tata_::available() {
	
	if ( message_ready && message_type == ASCII) {
		
		return word_count;
	}
  return 0 ;
  
}

uint8_t Tata_::sysexAvailable() {
	
	if ( message_ready && message_type == SYSEX) {
		
		return msg_length - current;
	}
  return 0 ;
  
}

uint8_t Tata_::sysexRead() {
	uint8_t value =0;
	if ( sysexAvailable() ) {
		
		value = buffer[current];
		current++;
		if (current >= msg_length) clear();
	}
	return value;
}

int Tata_::readInt() {
  int value = 0;
   
   if ( available() ) {
	   value = atoi(&buffer[current]);
	   nextWord();
	}
   

  return value;

}

// REPLACED ASCII METHOD WITH OUTPUT ONE CHAR BY CHAR
char Tata_::readChar() {
	 char value = 0;
	 if ( available() ) {
		 
		char value = buffer[current];
		if ( value == 0 ) value = ASCII_SEPARATOR;
		 current = current +1;
		 if ( current >= msg_length) clear();
	 }
  
  return value;

}


long Tata_::readLong() {
  if ( available()) { 
		long value = atol(&buffer[current]); // atol for long instead of atoi for int variables
		nextWord();
		return value; 
	}
	return 0;

}






void Tata_::copyString(char *string, uint8_t size) {
	
	if ( available() ) {
		strlcpy(string,&buffer[current],size);
		nextWord();
	} else {
		if ( size ) string[0] = '\0';
	}
}




uint8_t Tata_::checkString(char *string) {
	
	if ( available() && strcmp(string,&buffer[current]) == 0 ) {
			nextWord();
			return 1;
	}
	return 0;
}



void Tata_::nextWord() {

  if ( message_ready ) {
	  
		  while ( current < msg_length ) {
			  current++;
			  if ( buffer[current] == 0 ) {
				  word_count--;
				  current++;
				  if ( current < msg_length ) return;
				  break;
			  }
		  }
	  
	
	clear();
  }
  
 
}


void Tata_::printUnknown() {
	
	Serial.print("unknown ");
	if ( message_type == FIRMATA ) {
		// Print the unknown command
	  Serial.print(buffer[0]+buffer[1],DEC);
	} else {
		// Print the current buffer
	  while ( available() ) {
	      Serial.print((char *)(&buffer[current]));	
	      Serial.print(32,BYTE);
	      nextWord();
       }
       Serial.println();
	}
}






void Tata_::process() {
	
   
    while ( Serial.available( ) ) {
		
		uint8_t serialByte = Serial.read();
		
		if ( serialByte > 127 ) {
			
		
			
			if ( serialByte == END_SYSEX )  {
				if ( bufferIndex > 0 && callback != NULL && message_type == SYSEX ) {
					msg_length = bufferIndex;
					message_ready = 1;
					(*callback)() ;  
                }
				clear();
				
			} else {
			
				clear(); // Always clear the bugger once a FIRMATA command is received
				
				if ( serialByte == START_SYSEX ) {
					message_type = SYSEX;
					
				} else {
				
					if(serialByte < 0xF0) {
						buffer[0] = serialByte & 0xF0;
						buffer[1] = serialByte & 0x0F;
					} else {
						// commands in the 0xF* range don't use channel data
						buffer[0] = serialByte;
						buffer[1] = 0;
					}
					
					message_type = FIRMATA;
					bufferIndex=2;
					
					switch (buffer[0]) {
						case ANALOG_MESSAGE:
						case DIGITAL_MESSAGE:
						case SET_PIN_MODE:
						  msg_length = 2; // two data bytes needed
						  break;
						case REPORT_ANALOG:
						case REPORT_DIGITAL:
						  msg_length = 1; // one data bytes needed
						  break;
						case SYSTEM_RESET:
						  reset();
						  break;
						case REPORT_VERSION:
						  printVersionFirmata();
						  break;
						default: 
						  printUnknown();
						  clear();
					}
				}
			}
		} else {
			// SYSEX DATA
			if ( message_type == SYSEX ) {
				
				if ( bufferIndex == 0) {
					buffer[bufferIndex] = serialByte;
					sysex_MSB = 0;
					bufferIndex++;
				} else {
					if ( sysex_MSB ) {
						sysex_MSB = 0;
						buffer[bufferIndex] = buffer[bufferIndex] + (serialByte << 7);
						bufferIndex++;
					} else {
						buffer[bufferIndex] = serialByte;
						sysex_MSB = 1;
					}
				}
				
				 if (bufferIndex >= TATA_BUFFER_LAST_INDEX) clear();
			// FIRAMTA DATA
			} else if ( message_type == FIRMATA ) {
				msg_length--;
                buffer[bufferIndex] = serialByte;
                bufferIndex++;
                if ( msg_length==0 ) {
					doTheFIRMATA();
					clear();
				}
			// ASCII DATA	
			} else {
				
				if ( message_ready) clear();
     
			
				if ( serialByte == 13  )  { // CR
					if ( valid_char = 1 ) word_count++;
					buffer[bufferIndex]=0;
					msg_length = bufferIndex;
					
					if ( msg_length ) {
						message_ready = 1;
						doTheASCII();
						return;
					}
				} else if ( serialByte == ASCII_SEPARATOR) {
					if ( valid_char == 1 ) {
						word_count++;
						buffer[bufferIndex]=0;
					}
					valid_char = 0;
					bufferIndex++;
					if (bufferIndex >= TATA_BUFFER_LAST_INDEX) clear();
				} else if ( serialByte > 31 && serialByte < 128 ) {
					 buffer[bufferIndex]=serialByte;
					 bufferIndex++;
					 valid_char = 1;
					 if (bufferIndex >= TATA_BUFFER_LAST_INDEX) clear();
				}
			}
		}
		
		
	}
     
     // REPORTING
     // A call to process() must be done as often as possible.
     // Every call to process output at most one digital port and one analogPin.
     // The next output will be delayed as we wait for the message to be received.
     // A catchup mechanism should probably be implemented (if there is more than 
     // a millisecond between two calls to process().
     
   if(millis() >= millisToWait) {  
      int waitTime = 0;
      // find a digital pin to report
      uint8_t iterations;
      
      for ( iterations =0; iterations < TOTAL_PORTS; iterations++ ) {
		  reportDigitalIndex = (reportDigitalIndex + 1) % TOTAL_PORTS;
		  if ( reportDigital[reportDigitalIndex] ) {
			 
			  uint8_t value = readPort(reportDigitalIndex);
			  if ( value !=  reportDigitalPrevious[reportDigitalIndex] ) {
				  sendDigitalMessage(reportDigitalIndex,value);
				  waitTime++;
				  reportDigitalPrevious[reportDigitalIndex] = value;
				  break;
			}
		 }
	  }
      
      
      // find an analog pin to report
      for ( iterations =0; iterations < TOTAL_ANALOG_PINS; iterations++ ) {
		  reportAnalogIndex = (reportAnalogIndex + 1) %  TOTAL_ANALOG_PINS ;
		   if( reportAnalog[reportAnalogIndex / 8] & (1 << (reportAnalogIndex % 8)) ) {
				Serial.print(ANALOG_MESSAGE | (reportAnalogIndex & 0xF), BYTE);
				int value = analogRead(reportAnalogIndex);
				 Serial.print(value & B01111111, BYTE); // LSB
				 Serial.print(value >> 7 & B01111111, BYTE); // MSB
				 waitTime++;
				 break;
		  }
	  }
      
     millisToWait = millis() + waitTime;
   }
      
 }


void Tata_::sendDigitalMessage(uint8_t port, uint8_t value) {
 
    Serial.print(DIGITAL_MESSAGE | (port & 0xF),BYTE);
    Serial.print((byte)value % 128, BYTE); // Tx bits 0-6
    Serial.print(value >> 7, BYTE);  // Tx bits 7-13
    //previousPINs[port] = value;
    
 
}



void Tata_::doTheFIRMATA() {
	int value;
	// buffer[0] == MESSAGE_TYPE
	// buffer[1] == MESSAGE_CHANNEL/PORT
	// buffer[2] == DATA LSB/DISABLE-ENABLE
	// buffer[3] == DATA MSB
	
	 switch(buffer[0]) {
		case ANALOG_MESSAGE:
		   value = (buffer[3] << 7) + buffer[2];
			analogWrite(buffer[1], value);
			break;
						
		  case DIGITAL_MESSAGE:
		       value = (buffer[3] << 7) + buffer[2];
		       // This is a big limitation on the Firmata protocol
		       // It can change the state of pins we simply don't want changed
		       // StandardFirmata gets around this with a big array of pinMode management
		       // For now, we will deprecate it and create a  SINGLE_DIGITAL_MESSAGE
		       if (buffer[1] < TOTAL_PORTS) {
				   // We are disabling the pinMode mask with 0xFF
				   
				   writePort(buffer[1], (byte)value, 0xFF); 
			   }
		       
				break;
		  case SET_PIN_MODE:
		    // Exclude rx/tx pins
		    if ( buffer[2] > 1 ) pinMode(buffer[2],buffer[3]);
		     
			break;
		  case REPORT_ANALOG:
		   value = buffer[1] / 8; // find the analog port
		   if ( value < TOTAL_ANALOG_PORTS ) {
				if(buffer[2]) { // everything but 0 enables reporting of that pin
					reportAnalog[value] = reportAnalog[value] | (1 << (buffer[1]%8));
				} else { 
					reportAnalog[value] = reportAnalog[value] &~ (1 << (buffer[1]%8));
				}
			}
			break;
		  case REPORT_DIGITAL:
		    if ( buffer[1] < TOTAL_PORTS ) {
				reportDigital[buffer[1]] = buffer[2];
				
			}
			break;
		  }
	
}


void Tata_::reset() {
	// Copied from Bitlash
	// The following line is recommended but does not work on Arduino
	// Reset_AVR();
	// Instead we use this:
	void (*bootvec)(void) = 0; (*bootvec)(); 	// we jump through 0 instead
	
}


void Tata_::doTheASCII() {
	
  int pin;
  int state;
 
 #ifdef DEBUG_ASCII
 Serial.print("word count: ");
 Serial.println(word_count,DEC);
 #endif

  // This could probably be sped up
 
  // PINMODE
  if ( checkString("pm") ) { 
    while ( available() > 1 ) {      
        pinMode(readInt(),readInt());
    }
    return;
  } 

  // DIGITAL WRITE
  if ( checkString("dw") ) { 
    while ( available() > 1 ) {
      digitalWrite(readInt(),readInt());
    }
    return;
  } 

  // DIGITAL READ
  if ( checkString("dr") ) { 
    Serial.print("dr");
    while ( available() ) printInt(digitalRead(readInt()));
    Serial.println();
    return;
  } 

  // ANALOG READ
  if ( checkString("ar") ) { 
    Serial.print("ar");
    while ( available() ) printInt(analogRead(readInt()));
    Serial.println();
    return;
  }

  // ANALOG WRITE
  if ( checkString("aw") ) { 
    while ( available() > 1 ) {
      analogWrite(readInt(),readInt());
    }
    return;
  }
  
  // REBOOT
  if ( checkString("reset") ) { 
	reset(); 
	return;
  }
  
  if ( checkString("version") ) {
	  printVersion();
	  return;
	  
  }
  
   if ( callback != NULL) {
	  (*callback)() ;  
  }
  
 if ( available() ) printUnknown();
  
  

	
}


/*==============================================================================
 * readPort() - Read an 8 bit port
 *============================================================================*/


uint8_t Tata_::readPort(uint8_t port) {
#if defined(ARDUINO_PINOUT_OPTIMIZE)
	if (port == 0) return PIND & B11111100; // ignore Rx/Tx 0/1
	if (port == 1) return PINB & B00111111; // pins 8-13 (14,15 are disabled for the crystal)
	if (port == 2) return PINC;
	return 0;
#else
	unsigned char out=0, pin=port*8;
	if (IS_PIN_DIGITAL(pin+0) && digitalRead(PIN_TO_DIGITAL(pin+0))) out |= 0x01;
	if (IS_PIN_DIGITAL(pin+1) && digitalRead(PIN_TO_DIGITAL(pin+1))) out |= 0x02;
	if (IS_PIN_DIGITAL(pin+2) && digitalRead(PIN_TO_DIGITAL(pin+2))) out |= 0x04;
	if (IS_PIN_DIGITAL(pin+3) && digitalRead(PIN_TO_DIGITAL(pin+3))) out |= 0x08;
	if (IS_PIN_DIGITAL(pin+4) && digitalRead(PIN_TO_DIGITAL(pin+4))) out |= 0x10;
	if (IS_PIN_DIGITAL(pin+5) && digitalRead(PIN_TO_DIGITAL(pin+5))) out |= 0x20;
	if (IS_PIN_DIGITAL(pin+6) && digitalRead(PIN_TO_DIGITAL(pin+6))) out |= 0x40;
	if (IS_PIN_DIGITAL(pin+7) && digitalRead(PIN_TO_DIGITAL(pin+7))) out |= 0x80;
	return out;
#endif
}

/*==============================================================================
 * writePort() - Write an 8 bit port, only touch pins specified by a bitmask
 *============================================================================*/

uint8_t Tata_::writePort(uint8_t port, uint8_t value, uint8_t bitmask)
{
#if defined(ARDUINO_PINOUT_OPTIMIZE)
	if (port == 0) {
		bitmask = bitmask & 0xFC;  // Tx & Rx pins
		cli();
		PORTD = (PORTD & ~bitmask) | (bitmask & value);
		sei();
	} else if (port == 1) {
		cli();
		PORTB = (PORTB & ~bitmask) | (bitmask & value);
		sei();
	} else if (port == 2) {
		cli();
		PORTC = (PORTC & ~bitmask) | (bitmask & value);
		sei();
	}
#else
	byte pin=port*8;
	if ((bitmask & 0x01)) digitalWrite(PIN_TO_DIGITAL(pin+0), (value & 0x01));
	if ((bitmask & 0x02)) digitalWrite(PIN_TO_DIGITAL(pin+1), (value & 0x02));
	if ((bitmask & 0x04)) digitalWrite(PIN_TO_DIGITAL(pin+2), (value & 0x04));
	if ((bitmask & 0x08)) digitalWrite(PIN_TO_DIGITAL(pin+3), (value & 0x08));
	if ((bitmask & 0x10)) digitalWrite(PIN_TO_DIGITAL(pin+4), (value & 0x10));
	if ((bitmask & 0x20)) digitalWrite(PIN_TO_DIGITAL(pin+5), (value & 0x20));
	if ((bitmask & 0x40)) digitalWrite(PIN_TO_DIGITAL(pin+6), (value & 0x40));
	if ((bitmask & 0x80)) digitalWrite(PIN_TO_DIGITAL(pin+7), (value & 0x80));
#endif
}




void Tata_::printInt(int data) {
  Serial.print(' ',BYTE);
  Serial.print(data);
}

// Preinstantiate Object //////////////////////////////////////////////////////
Tata_ Tata = Tata_();
